% Text Mining with MATLAB 2nd Edition
% www.http://www.textmininglab.net/

% CHAPTER 15: Question Answering and Dialogue

disp('Example 15.1 -----------------------------------------------------');
clear; load datasets_ch15; whos                                    %(15.1a)
fieldnames(data)'                                                  %(15.1b)
pause;

disp('Example 15.2 -----------------------------------------------------');
data(1)                                                             %(15.2)
pause;

disp('Example 15.3 -----------------------------------------------------');
datatbl = struct2table(data);                                      %(15.3a)
restaurants = unique(datatbl.restaurant);                          %(15.3b)
cuisines = setdiff(unique(datatbl.cuisine),"");
prices = unique(datatbl.price);
locations = unique(datatbl.location);
whos restaurants cuisines prices locations                         %(15.3c)
pause;

disp('Example 15.4 -----------------------------------------------------');
question = "What kind of food is served at Sandrino Pizza?";        %(15.4)
intent = get_intent(question)
pause;

disp('Example 15.5 -----------------------------------------------------');
matches = property_match(restaurants,question)                      %(15.5)
pause;

disp('Example 15.6 -----------------------------------------------------');
restaurantname = matches.value(1);                                  %(15.6)
idx = datatbl.restaurant==restaurantname;
result = datatbl.cuisine(idx,:);
result = result(strlength(result)>0) 
pause;

disp('Example 15.7 -----------------------------------------------------');
template = "%s specializes in %s food.";                            %(15.7)
if length(result)>1                                                
    result = join(result(1:end-1),', ')+" and "+result(end);
end
answer = compose(template,replace([restaurantname,result],'_',' '))
pause;

%%%%%%%% Table 15.2 to show various responses
temp1 = "What type of food is served at Barrel House?";
temp2 = "What is Agave's cuisine?";
temp3 = "Is Galeto Steakhouse a Japanese restaurant?";
temp4 = "Which type of cuisine is there at Sartaj Cafe?";
temp5 = "What cuisine does Zazie restaurant specializes in?";
queries = [temp1;temp2;temp3;temp4;temp5];

for k=1:length(queries)
    disp(queries(k));
    matches = property_match(restaurants,queries(k));
    restaurantname = matches.value(1);
    idx = datatbl.restaurant==restaurantname;
    result = datatbl.cuisine(idx,:);
    result = result(strlength(result)>0);
    if length(result)>1
        result = join(result(1:end-1),', ')+" and "+result(end);
    end
    template = "%s specializes in %s food.\n";
    disp(compose(template,replace([restaurantname,result],'_',' ')));
end
pause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 15.8 -----------------------------------------------------');
templatetbl = table();                                              %(15.8)
templatetbl.type = ["cuisine"; "location"; "price"; "website"; ...
                    "address"; "telephone"];
templatetbl.template = ["%s specializes in %s food."; ...
                        "%s is located in %s."; ...
                        "%s price range is %s."; ...
                        "The website of %s is %s"; ...
                        "The address of %s is %s."; ...
                        "%s can be reached at %s"];
pause;

disp('Example 15.9 -----------------------------------------------------');
question = "Where is Sandrino's Pizza located?";                   %(15.9a)
intent = get_intent(question);
matches = property_match(restaurants,question);                   
restaurantname = matches.value(1);                                 %(15.9b)
idx = datatbl.restaurant==restaurantname;
searchfield = split(intent,'_');
searchfield = searchfield(end);
result = datatbl.(searchfield)(idx,:);
result = result(strlength(result)>0);
templateidx = templatetbl.type==searchfield;                       %(15.9c)
template = templatetbl.template(templateidx);
if length(result)>1                                                
    result = join(result(1:end-1),', ')+" and "+result(end);
end
answer = compose(template,replace([restaurantname,result],'_',' '))
pause;

%%%%%%%% Table 15.3 to show various responses
temp1 = "What city is Marufuku located at?";
temp2 = "How are the prices at Scott Grill?";
temp3 = "Is there a contact number for Nopa?";
temp4 = "What is the address of waterbar?";
temp5 = "Is there a website for catch seafood?";
queries = [temp1;temp2;temp3;temp4;temp5];

for k=1:length(queries)
    disp(queries(k));
    
    intent = get_intent(queries(k));
    matches = property_match(restaurants,queries(k));                   
    restaurantname = matches.value(1); 
    idx = datatbl.restaurant==restaurantname;
    searchfield = split(intent,'_');
    searchfield = searchfield(end);
    result = datatbl.(searchfield)(idx,:);
    result = result(strlength(result)>0);
    templateidx = templatetbl.type==searchfield;  
    template = templatetbl.template(templateidx);
    if length(result)>1                                                
        result = join(result(1:end-1),', ')+" and "+result(end);
    end
    disp(compose(template+"\n",replace([restaurantname,result],'_',' ')))

end
pause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


disp('Example 15.10 ----------------------------------------------------');
question = "Is there any japanese restaurant in Oakland?";        %(15.10a)
locationvalue = property_match(locations,question)
locationidx = datatbl.location==locationvalue.value(1);           %(15.10b)
disp(sum(locationidx))
pause;

disp('Example 15.11 ----------------------------------------------------');
cuisinetype = property_match(cuisines,question)                   %(15.11a)
cuisineidx = sum(datatbl.cuisine==cuisinetype.value(1),2);        %(15.11b)
disp(sum(cuisineidx))
pause;

disp('Example 15.12 ----------------------------------------------------');
intent = get_intent(question)                                     %(15.12a)
datatbl.restaurant(locationidx&cuisineidx)'                       %(15.12b)
pause;

disp('Example 15.13 ----------------------------------------------------');
pricerange = property_match(prices,question)                       %(15.13)
pause;

disp('Example 15.14 ----------------------------------------------------');
constraints(1).name = "cuisine";                                  %(15.14a)
constraints(1).values = table2struct(cuisinetype);
constraints(2).name = "location";
constraints(2).values = table2struct(locationvalue);
constraints(3).name = "price";                                    
constraints(3).values = table2struct(pricerange);
searchindex = find_restaurant(datatbl,constraints);               %(15.14b)
datatbl.restaurant(searchindex)'
pause;


disp('Example 15.15 ----------------------------------------------------');
question = "What is the location of La Ciccia Restaurant?";       %(15.15a)
intent = get_intent(question);
matches = property_match(restaurants,question); 
newconstraint(1).name = "restaurant";                             %(15.15b)
newconstraint(1).values = table2struct(matches);
searchindex = find_restaurant(datatbl,newconstraint);
responsefield = split(intent,'_');
datatbl.(responsefield(end))(searchindex,:)
pause;

disp('Example 15.16 ----------------------------------------------------');
question = "Is there any spanish restaurant with a " + ...        %(15.16a)
           "moderate price range?";
intent = get_intent(question); % detects the intent               %(15.16b)
% looks for all possible informable variable mentions             %(15.16c)
variables = ["cuisine","price","location"];
valuelists = {cuisines,prices,locations};
constraints = struct();
for k=1:length(variables)
    matches = property_match(valuelists{k},question);
    constraints(k).name = variables(k);
    constraints(k).values = table2struct(matches);
end
pause;

disp('Example 15.17 ----------------------------------------------------');
searchindex = find_restaurant(datatbl,constraints);                %(15.17)
responsefield = split(intent,'_');
responsefield = responsefield(end);
result = datatbl.(responsefield)(searchindex,:);
result = result(strlength(result)>0);
if length(result)>1
    result = join(result(1:end-1),', ')+" and "+result(end);
end
nresponses = sum(searchindex);
pause;

disp('Example 15.18 ----------------------------------------------------');
commontemplate = "_price_ _cuisine_ restaurants in _location_";    %(15.18)
if nresponses < 1
    template = "I didn't find any "+commontemplate+".";
elseif nresponses < 2
    template = "_results_ is a "+commontemplate+".";
elseif nresponses < 5
    template = "I found _number_ "+commontemplate+": _results_.";
else
    template = "I found _number_ "+commontemplate+".";
end
pause;

disp('Example 15.19 ----------------------------------------------------');
% replaces insertion patterns corresponding to results            %(15.19a)
template = replace(template,'_number_',string(nresponses)); 
if not(isempty(result))
    template = replace(template,'_results_',result);
end
% replaces insertion patterns corresponding to informable values  %(15.19b)
for k=1:length(variables)
    pattern = "_"+variables(k)+"_";
    if isempty(constraints(k).values), value = "";
    else, value = constraints(k).values(1).value; end
    template = replace(template,pattern,value);
end
answer = template                                                 %(15.19c)
pause;

disp('Example 15.20 ----------------------------------------------------');
answer = regexprep(answer,'_',' ');                               %(15.20a)
answer = regexprep(answer,' +',' ');
if nresponses==1
    answer = replace(answer,'restaurants','restaurant');
end
answer = replace(answer," in .",".");
answer = replace(answer," in :",":");
answer = replace(answer," a expensive "," an expensive ");
answer = replace(answer," moderate price "," moderate-price ");
disp(answer)                                                      %(15.20b)
pause;

%%%%%%%%%%%%% Table 15.4
% more than 4
q1 = "Are there any restaurants offering new american cuisine?"
% three options
q2 = "Are there any pizza places in Sausalito?"
% one option
q3 = "Is there any restaurant in Palo Alto serving steaks?"
% one option
q4 = "Is there any greek restaurant in Oakland?"
% zero options
q5 = "Is there an expensive Mexican restaurant in Palo Alto?"
queries = [q1,q2,q3,q4];

variables = ["cuisine","price","location"];
valuelists = {cuisines,prices,locations};

for k=1:length(queries)
disp(queries(k));
    
% nlu    
question = queries(k);
intent = get_intent(question);
constraints = struct();
for k=1:length(variables)
    matches = property_match(valuelists{k},question);
    constraints(k).name = variables(k);
    constraints(k).values = table2struct(matches);
end

% search
searchindex = find_restaurant(datatbl,constraints);
responsefield = split(intent,'_');
responsefield = responsefield(end);
result = datatbl.(responsefield)(searchindex,:);
result = result(strlength(result)>0);
if length(result)>1
    result = join(result(1:end-1),', ')+" and "+result(end);
end
nresponses = sum(searchindex);

% nlg (step 1)
commontemplate = "_price_ _cuisine_ restaurants in _location_"; 
if nresponses < 1
    template = "I didn't find any "+commontemplate+".";
elseif nresponses < 2
    template = "_results_ is a "+commontemplate+".";
elseif nresponses < 5
    template = "I found _number_ "+commontemplate+": _results_.";
else
    template = "I found _number_ "+commontemplate+".";
end

% nlg (step 2)
template = replace(template,'_number_',string(nresponses));       
if not(isempty(result))
    template = replace(template,'_results_',result);
end
for k=1:length(variables)
    pattern = "_"+variables(k)+"_";
    if isempty(constraints(k).values), value = "";
    else, value = constraints(k).values(1).value; end
    template = replace(template,pattern,value);
end
answer = template;

% nlg (step 3)
answer = regexprep(answer,'_',' ');
answer = regexprep(answer,' +',' ');
if nresponses==1
    answer = replace(answer,'restaurants','restaurant');
end
answer = replace(answer," in .",".");
answer = replace(answer," in :",":");
answer = replace(answer," a expensive "," an expensive ");
answer = replace(answer," moderate price "," moderate-price ");

disp(sprintf(answer+"\n"));
end
pause;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 15.21 ----------------------------------------------------');
disp('Not a script');                                              %(15.21)
pause;

disp('Example 15.22 ----------------------------------------------------');
get_reward(4,["Chinese","","Expensive",""])                       %(15.22a)
get_reward(2,["","Palo_Alto","",""])                              %(15.22b)
get_reward(3,["Mediterranean","","",""])                          %(15.22c)
pause;

disp('Example 15.23 ----------------------------------------------------');
system_response(1,["","","Expensive",""])                         %(15.23a)
system_response(4,["","San_Francisco","",""])                     %(15.23b)
pause;

disp('Example 15.24 ----------------------------------------------------');
rng('default');                                                   %(15.24a)
systemquery = "What city are you looking for?";
[value,response,goals] = user_response(systemquery,struct())
systemquery = "I found 1 restaurant!";                            %(15.24b)
[value,response,goals] = user_response(systemquery,goals)
pause;

disp('Example 15.25 ----------------------------------------------------');
% initializes the training session                                %(15.25a)
maxepisodes = 50; maxturns = 20;
Qmtx = zeros(16,4); finalstate = 16;
% reinforcement learning training session                         %(15.25b)
for episode = 1:maxepisodes                           
    % initializes a new dialogue session
    slots = ["","","",""]; goals = struct();
    cumreward = 0; dialogue = "";
    for k=1:maxturns % simulates a new dialogue session
        state = (strlength(slots)>0)*[1;2;4;8]+1;
        if state==finalstate, break; end
        % selects an action randomly from max Qmtx values
        possibleactions = find(Qmtx(state,:)==max(Qmtx(state,:)));
        idx = randperm(length(possibleactions));
        action = possibleactions(idx(1));
        % computes the system response and updates the dialogue log
        systemresponse = system_response(action,slots);
        dialogue = dialogue + "System: " + systemresponse + newline;
        % computes the immediate reward and updated the cumulative one
        reward = get_reward(action,slots);
        cumreward = cumreward + reward;
        % computes the simulated user response and updates the log
        [value,userresponse,goals] = user_response(systemresponse,goals);
        dialogue = dialogue + "  User: " + userresponse + newline;
        % fills in the corresponding slot and computes the new state
        slots(action) = value;
        newstate = (strlength(slots)>0)*[1;2;4;8]+1;   
        % implements the Q-learning step
        alpha = 0.7; beta = 0.9;
        Qmtx(state,action) = (1-alpha)*Qmtx(state,action)+...
                             alpha*(reward+beta*max(Qmtx(newstate,:)));
    end
    % saves relevant info about the concluded dialogue session
    rewards(episode) = cumreward;
    turns(episode) = k-1;
    Qmtxs{episode} = Qmtx;
    dialogues{episode} = dialogue;
end
pause;

disp('Example 15.26 ----------------------------------------------------');
hf = figure(5);                                                    %(15.26)
set(hf,'Color',[1 1 1],'Name','Reinforcement Learning Training Session');
subplot(2,1,1); plot(turns,'.-');
ylabel('Number of dialogue turns')
xlabel('Number of simulated dialogue sessions (episodes)')
subplot(2,1,2); plot(rewards,'.-');
ylabel('Overall accumulated reward')
xlabel('Number of simulated dialogue sessions (episodes)')
pause;

disp('Example 15.27 ----------------------------------------------------');
finalQmtx = Qmtxs{maxepisodes};                                   %(15.27a)
policy = zeros(8,4);
for k=1:8 
    idx = find(finalQmtx(k,:)==max(finalQmtx(k,:)));
    policy(k,idx) = 1/length(idx);
end
% writes the learned dialogue policy into a table                 %(15.27b)
policytbl = array2table(policy,'VariableNames',compose('Action %d',1:4));
policytbl = addvars(policytbl,compose("State %d",1:8)',...
                    'Before','Action 1','NewVariableNames','State')
pause;

disp('Example 15.28 ----------------------------------------------------');
disp(dialogues{1})                                                 %(15.28)
pause;

disp('Example 15.29 ----------------------------------------------------');
disp(dialogues{15})                                                %(15.29)
pause;

disp('Example 15.30 ----------------------------------------------------');
disp(dialogues{20})                                               %(15.30a)
disp(dialogues{30})                                               %(15.30b)
disp(dialogues{45})                                               %(15.30c)
pause;

% GENERATES FIGURE 15.6
Qmtxs = [zeros(16,4),Qmtxs];

rng('default'); 
for stage = 1:length(Qmtxs)
    mtx = Qmtxs{stage};
    rwds= [];
    trns= [];

for sim = 1:10000
    slots = ["","","",""]; % food, location, price, selection   
    cumreward = 0;
    goals = struct();
    
    for k=1:100 
        state = (strlength(slots)>0)*[1;2;4;8]+1;
        if state==finalstate, break; end
        
        % selects actions randomly from max values
        possibleactions = find(mtx(state,:)==max(mtx(state,:)));
        idx = randperm(length(possibleactions));
        action = possibleactions(idx(1));
        
        %reward = rewardmtx(state,action);
        reward = get_reward(action,slots);
        cumreward = cumreward + reward;
        
        slots(action) = "x";
        
    end
    rwds(sim)=cumreward;
    trns(sim)=k-1;
    
end

    mrwds(stage) = mean(rwds);
    hrwds(stage) = max(rwds);
    lrwds(stage) = min(rwds);
    mtrns(stage) = mean(trns);
    htrns(stage) = max(trns);
    ltrns(stage) = min(trns);
end

hf = figure(6);
set(hf,'Color',[1 1 1],'Name','Averaged Turns and Rewards');
subplot(2,1,2);
plot(0:50,mrwds,'.-');
ylabel('Average overall reward')
xlabel('Learning episodes')
subplot(2,1,1);
plot(0:50,mtrns,'.-');
ylabel('Average number of turns')
xlabel('Learning episodes')


